package garden_integration_tests_test

import (
	"fmt"
	"net"
	"os/exec"
	"strings"
	"time"

	"github.com/cloudfoundry-incubator/garden"
	"github.com/eapache/go-resiliency/retrier"
	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/onsi/gomega/gbytes"
	"github.com/onsi/gomega/gexec"
)

var _ = Describe("Networking", func() {
	It("can be contacted after a NetIn", func() {
		_, err := container.Run(garden.ProcessSpec{
			Path: "sh",
			Args: []string{"-c", "echo hallo | nc -l -p 8080"},
			User: "root",
		}, garden.ProcessIO{
			Stdout: GinkgoWriter,
			Stderr: GinkgoWriter,
		})
		Expect(err).ToNot(HaveOccurred())

		hostPort, _, err := container.NetIn(0, 8080)
		Expect(err).ToNot(HaveOccurred())

		gardenHostname := strings.Split(gardenHost, ":")[0]
		Eventually(func() error {
			nc, err := gexec.Start(exec.Command("nc", gardenHostname, fmt.Sprintf("%d", hostPort)), GinkgoWriter, GinkgoWriter)
			if err != nil {
				Eventually(nc).Should(gbytes.Say("hallo"))
			}

			return err
		}).ShouldNot(HaveOccurred())
	})

	Describe("domain name resolution", func() {
		itCanResolve := func(domainName string) {
			defer func() {
				err := gardenClient.Destroy(container.Handle())
				Expect(err).NotTo(HaveOccurred())
			}()
			output := gbytes.NewBuffer()

			err := retrier.New(retrier.ConstantBackoff(30, 2*time.Second), nil).Run(func() error {
				proc, err := container.Run(garden.ProcessSpec{
					// We are using ping here rather than nslookup as we saw some
					// flakey behaviour with nslookup on our local concourse machines.
					// We're testing on the output of ping, which reports "bad address"
					// if it is unable to resolve a domain.
					Path: "ping",
					Args: []string{"-c", "1", domainName},
					User: "root",
				}, garden.ProcessIO{Stdout: output, Stderr: output})
				Expect(err).NotTo(HaveOccurred())

				_, err = proc.Wait()
				return err
			})

			Expect(err).NotTo(HaveOccurred())
			Expect(output).ToNot(gbytes.Say("ping: bad address"))
		}

		It("can resolve localhost", func() {
			itCanResolve("localhost")
		})

		It("can resolve its hostname", func() {
			itCanResolve(container.Handle())
		})

		It("can resolve domain names", func() {
			itCanResolve("www.example.com")
		})
	})
})

func checkInternet(container garden.Container, externalIP net.IP) error {
	return checkConnection(container, externalIP.String(), 80)
}

func checkConnection(container garden.Container, ip string, port int) error {
	process, err := container.Run(garden.ProcessSpec{
		User: "alice",
		Path: "sh",
		Args: []string{"-c", fmt.Sprintf("echo hello | nc -w1 %s %d", ip, port)},
	}, garden.ProcessIO{Stdout: GinkgoWriter, Stderr: GinkgoWriter})
	if err != nil {
		return err
	}

	exitCode, err := process.Wait()
	if err != nil {
		return err
	}

	if exitCode == 0 {
		return nil
	} else {
		return fmt.Errorf("Request failed. Process exited with code %d", exitCode)
	}
}
